/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.hadoop.ozone.s3.endpoint;

import static javax.ws.rs.core.HttpHeaders.CONTENT_LENGTH;
import static javax.ws.rs.core.HttpHeaders.ETAG;
import static javax.ws.rs.core.HttpHeaders.LAST_MODIFIED;
import static org.apache.hadoop.hdds.protocol.proto.HddsProtos.ReplicationType.EC;
import static org.apache.hadoop.hdds.scm.ScmConfigKeys.OZONE_SCM_CHUNK_SIZE_DEFAULT;
import static org.apache.hadoop.hdds.scm.ScmConfigKeys.OZONE_SCM_CHUNK_SIZE_KEY;
import static org.apache.hadoop.ozone.OzoneConfigKeys.HDDS_CONTAINER_RATIS_DATASTREAM_ENABLED;
import static org.apache.hadoop.ozone.OzoneConfigKeys.HDDS_CONTAINER_RATIS_DATASTREAM_ENABLED_DEFAULT;
import static org.apache.hadoop.ozone.OzoneConfigKeys.OZONE_FS_DATASTREAM_AUTO_THRESHOLD;
import static org.apache.hadoop.ozone.OzoneConfigKeys.OZONE_FS_DATASTREAM_AUTO_THRESHOLD_DEFAULT;
import static org.apache.hadoop.ozone.OzoneConfigKeys.OZONE_REPLICATION;
import static org.apache.hadoop.ozone.OzoneConfigKeys.OZONE_REPLICATION_TYPE;
import static org.apache.hadoop.ozone.OzoneConfigKeys.OZONE_REPLICATION_TYPE_DEFAULT;
import static org.apache.hadoop.ozone.audit.AuditLogger.PerformanceStringBuilder;
import static org.apache.hadoop.ozone.s3.S3GatewayConfigKeys.OZONE_S3G_CLIENT_BUFFER_SIZE_DEFAULT;
import static org.apache.hadoop.ozone.s3.S3GatewayConfigKeys.OZONE_S3G_CLIENT_BUFFER_SIZE_KEY;
import static org.apache.hadoop.ozone.s3.S3GatewayConfigKeys.OZONE_S3G_FSO_DIRECTORY_CREATION_ENABLED;
import static org.apache.hadoop.ozone.s3.S3GatewayConfigKeys.OZONE_S3G_FSO_DIRECTORY_CREATION_ENABLED_DEFAULT;
import static org.apache.hadoop.ozone.s3.exception.S3ErrorTable.ENTITY_TOO_SMALL;
import static org.apache.hadoop.ozone.s3.exception.S3ErrorTable.INVALID_ARGUMENT;
import static org.apache.hadoop.ozone.s3.exception.S3ErrorTable.INVALID_REQUEST;
import static org.apache.hadoop.ozone.s3.exception.S3ErrorTable.NOT_IMPLEMENTED;
import static org.apache.hadoop.ozone.s3.exception.S3ErrorTable.NO_SUCH_UPLOAD;
import static org.apache.hadoop.ozone.s3.exception.S3ErrorTable.PRECOND_FAILED;
import static org.apache.hadoop.ozone.s3.exception.S3ErrorTable.newError;
import static org.apache.hadoop.ozone.s3.util.S3Consts.ACCEPT_RANGE_HEADER;
import static org.apache.hadoop.ozone.s3.util.S3Consts.CONTENT_RANGE_HEADER;
import static org.apache.hadoop.ozone.s3.util.S3Consts.COPY_SOURCE_HEADER;
import static org.apache.hadoop.ozone.s3.util.S3Consts.COPY_SOURCE_HEADER_RANGE;
import static org.apache.hadoop.ozone.s3.util.S3Consts.COPY_SOURCE_IF_MODIFIED_SINCE;
import static org.apache.hadoop.ozone.s3.util.S3Consts.COPY_SOURCE_IF_UNMODIFIED_SINCE;
import static org.apache.hadoop.ozone.s3.util.S3Consts.CUSTOM_METADATA_COPY_DIRECTIVE_HEADER;
import static org.apache.hadoop.ozone.s3.util.S3Consts.CopyDirective;
import static org.apache.hadoop.ozone.s3.util.S3Consts.DECODED_CONTENT_LENGTH_HEADER;
import static org.apache.hadoop.ozone.s3.util.S3Consts.MP_PARTS_COUNT;
import static org.apache.hadoop.ozone.s3.util.S3Consts.RANGE_HEADER;
import static org.apache.hadoop.ozone.s3.util.S3Consts.RANGE_HEADER_SUPPORTED_UNIT;
import static org.apache.hadoop.ozone.s3.util.S3Consts.STORAGE_CLASS_HEADER;
import static org.apache.hadoop.ozone.s3.util.S3Consts.TAG_COUNT_HEADER;
import static org.apache.hadoop.ozone.s3.util.S3Consts.TAG_DIRECTIVE_HEADER;
import static org.apache.hadoop.ozone.s3.util.S3Utils.urlDecode;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.ImmutableMap;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.security.DigestInputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.ParseException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.OptionalLong;
import javax.annotation.PostConstruct;
import javax.inject.Inject;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.HEAD;
import javax.ws.rs.HeaderParam;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.container.ContainerRequestContext;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.MultivaluedMap;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.StreamingOutput;
import javax.xml.bind.DatatypeConverter;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.hadoop.hdds.client.ECReplicationConfig;
import org.apache.hadoop.hdds.client.ReplicationConfig;
import org.apache.hadoop.hdds.client.ReplicationType;
import org.apache.hadoop.hdds.conf.OzoneConfiguration;
import org.apache.hadoop.hdds.conf.StorageUnit;
import org.apache.hadoop.ozone.OzoneConsts;
import org.apache.hadoop.ozone.audit.S3GAction;
import org.apache.hadoop.ozone.client.OzoneBucket;
import org.apache.hadoop.ozone.client.OzoneKey;
import org.apache.hadoop.ozone.client.OzoneKeyDetails;
import org.apache.hadoop.ozone.client.OzoneMultipartUploadPartListParts;
import org.apache.hadoop.ozone.client.OzoneVolume;
import org.apache.hadoop.ozone.client.io.OzoneInputStream;
import org.apache.hadoop.ozone.client.io.OzoneOutputStream;
import org.apache.hadoop.ozone.om.OmConfig;
import org.apache.hadoop.ozone.om.exceptions.OMException;
import org.apache.hadoop.ozone.om.exceptions.OMException.ResultCodes;
import org.apache.hadoop.ozone.om.helpers.BucketLayout;
import org.apache.hadoop.ozone.om.helpers.OmMultipartCommitUploadPartInfo;
import org.apache.hadoop.ozone.om.helpers.OmMultipartInfo;
import org.apache.hadoop.ozone.om.helpers.OmMultipartUploadCompleteInfo;
import org.apache.hadoop.ozone.s3.HeaderPreprocessor;
import org.apache.hadoop.ozone.s3.SignedChunksInputStream;
import org.apache.hadoop.ozone.s3.endpoint.S3Tagging.Tag;
import org.apache.hadoop.ozone.s3.exception.OS3Exception;
import org.apache.hadoop.ozone.s3.exception.S3ErrorTable;
import org.apache.hadoop.ozone.s3.util.RFC1123Util;
import org.apache.hadoop.ozone.s3.util.RangeHeader;
import org.apache.hadoop.ozone.s3.util.RangeHeaderParserUtil;
import org.apache.hadoop.ozone.s3.util.S3Consts;
import org.apache.hadoop.ozone.s3.util.S3StorageType;
import org.apache.hadoop.ozone.s3.util.S3Utils;
import org.apache.hadoop.ozone.web.utils.OzoneUtils;
import org.apache.hadoop.util.Time;
import org.apache.http.HttpStatus;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Key level rest endpoints.
 */
@Path("/{bucket}/{path:.+}")
public class ObjectEndpoint extends EndpointBase {

  private static final Logger LOG =
      LoggerFactory.getLogger(ObjectEndpoint.class);

  private static final ThreadLocal<MessageDigest> E_TAG_PROVIDER;

  static {
    E_TAG_PROVIDER = ThreadLocal.withInitial(() -> {
      try {
        return MessageDigest.getInstance(OzoneConsts.MD5_HASH);
      } catch (NoSuchAlgorithmException e) {
        throw new RuntimeException(e);
      }
    });
  }

  @Context
  private ContainerRequestContext context;

  @Context
  private HttpHeaders headers;

  /*FOR the feature Overriding Response Header
  https://docs.aws.amazon.com/de_de/AmazonS3/latest/API/API_GetObject.html */
  private Map<String, String> overrideQueryParameter;
  private int bufferSize;
  private int chunkSize;
  private boolean datastreamEnabled;
  private long datastreamMinLength;

  public ObjectEndpoint() {
    overrideQueryParameter = ImmutableMap.<String, String>builder()
        .put("Content-Type", "response-content-type")
        .put("Content-Language", "response-content-language")
        .put("Expires", "response-expires")
        .put("Cache-Control", "response-cache-control")
        .put("Content-Disposition", "response-content-disposition")
        .put("Content-Encoding", "response-content-encoding")
        .build();
  }

  @Inject
  private OzoneConfiguration ozoneConfiguration;

  @Override
  @PostConstruct
  public void init() {
    bufferSize = (int) ozoneConfiguration.getStorageSize(
        OZONE_S3G_CLIENT_BUFFER_SIZE_KEY,
        OZONE_S3G_CLIENT_BUFFER_SIZE_DEFAULT, StorageUnit.BYTES);
    chunkSize = (int) ozoneConfiguration.getStorageSize(
        OZONE_SCM_CHUNK_SIZE_KEY,
        OZONE_SCM_CHUNK_SIZE_DEFAULT,
        StorageUnit.BYTES);
    datastreamEnabled = ozoneConfiguration.getBoolean(
        HDDS_CONTAINER_RATIS_DATASTREAM_ENABLED,
        HDDS_CONTAINER_RATIS_DATASTREAM_ENABLED_DEFAULT);
    datastreamMinLength = (long) ozoneConfiguration.getStorageSize(
        OZONE_FS_DATASTREAM_AUTO_THRESHOLD,
        OZONE_FS_DATASTREAM_AUTO_THRESHOLD_DEFAULT, StorageUnit.BYTES);
  }

  /**
   * Rest endpoint to upload object to a bucket.
   * <p>
   * See: https://docs.aws.amazon.com/AmazonS3/latest/API/RESTObjectPUT.html for
   * more details.
   */
  @SuppressWarnings({"checkstyle:MethodLength", "checkstyle:ParameterNumber"})
  @PUT
  public Response put(
      @PathParam("bucket") String bucketName,
      @PathParam("path") String keyPath,
      @HeaderParam("Content-Length") long length,
      @QueryParam("partNumber")  int partNumber,
      @QueryParam("uploadId") @DefaultValue("") String uploadID,
      @QueryParam("tagging") String taggingMarker,
      @QueryParam("acl") String aclMarker,
      final InputStream body) throws IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();
    S3GAction s3GAction = S3GAction.CREATE_KEY;
    boolean auditSuccess = true;
    PerformanceStringBuilder perf = new PerformanceStringBuilder();

    String copyHeader = null, storageType = null;
    DigestInputStream digestInputStream = null;
    try {
      if (aclMarker != null) {
        s3GAction = S3GAction.PUT_OBJECT_ACL;
        throw newError(NOT_IMPLEMENTED, keyPath);
      }
      OzoneVolume volume = getVolume();
      if (taggingMarker != null) {
        s3GAction = S3GAction.PUT_OBJECT_TAGGING;
        return putObjectTagging(volume, bucketName, keyPath, body);
      }

      if (uploadID != null && !uploadID.equals("")) {
        if (headers.getHeaderString(COPY_SOURCE_HEADER) == null) {
          s3GAction = S3GAction.CREATE_MULTIPART_KEY;
        } else {
          s3GAction = S3GAction.CREATE_MULTIPART_KEY_BY_COPY;
        }
        // If uploadID is specified, it is a request for upload part
        return createMultipartKey(volume, bucketName, keyPath, length,
            partNumber, uploadID, body, perf);
      }

      copyHeader = headers.getHeaderString(COPY_SOURCE_HEADER);
      storageType = headers.getHeaderString(STORAGE_CLASS_HEADER);
      boolean storageTypeDefault = StringUtils.isEmpty(storageType);

      // Normal put object
      OzoneBucket bucket = volume.getBucket(bucketName);
      ReplicationConfig replicationConfig =
          getReplicationConfig(bucket, storageType);

      boolean enableEC = false;
      if ((replicationConfig != null &&
          replicationConfig.getReplicationType() == EC) ||
          bucket.getReplicationConfig() instanceof ECReplicationConfig) {
        enableEC = true;
      }

      if (copyHeader != null) {
        //Copy object, as copy source available.
        s3GAction = S3GAction.COPY_OBJECT;
        CopyObjectResponse copyObjectResponse = copyObject(volume,
            copyHeader, bucketName, keyPath, replicationConfig,
            storageTypeDefault, perf);
        return Response.status(Status.OK).entity(copyObjectResponse).header(
            "Connection", "close").build();
      }

      boolean canCreateDirectory = ozoneConfiguration
          .getBoolean(OZONE_S3G_FSO_DIRECTORY_CREATION_ENABLED,
              OZONE_S3G_FSO_DIRECTORY_CREATION_ENABLED_DEFAULT) &&
          bucket.getBucketLayout() == BucketLayout.FILE_SYSTEM_OPTIMIZED;

      String amzDecodedLength =
          headers.getHeaderString(S3Consts.DECODED_CONTENT_LENGTH_HEADER);
      boolean hasAmzDecodedLengthZero = amzDecodedLength != null &&
          Long.parseLong(amzDecodedLength) == 0;
      if (canCreateDirectory &&
          (length == 0 || hasAmzDecodedLengthZero) &&
          StringUtils.endsWith(keyPath, "/")
      ) {
        s3GAction = S3GAction.CREATE_DIRECTORY;
        getClientProtocol()
            .createDirectory(volume.getName(), bucketName, keyPath);
        long metadataLatencyNs =
            getMetrics().updatePutKeyMetadataStats(startNanos);
        perf.appendMetaLatencyNanos(metadataLatencyNs);
        return Response.ok().status(HttpStatus.SC_OK).build();
      }

      // Normal put object
      Map<String, String> customMetadata =
          getCustomMetadataFromHeaders(headers.getRequestHeaders());

      if (S3Utils.hasSignedPayloadHeader(headers)) {
        digestInputStream = new DigestInputStream(new SignedChunksInputStream(body),
            getMessageDigestInstance());
        length = Long.parseLong(amzDecodedLength);
      } else {
        digestInputStream = new DigestInputStream(body, getMessageDigestInstance());
      }

      Map<String, String> tags = getTaggingFromHeaders(headers);

      long putLength;
      String eTag = null;
      if (datastreamEnabled && !enableEC && length > datastreamMinLength) {
        perf.appendStreamMode();
        Pair<String, Long> keyWriteResult = ObjectEndpointStreaming
            .put(bucket, keyPath, length, replicationConfig, chunkSize,
                customMetadata, tags, digestInputStream, perf);
        eTag = keyWriteResult.getKey();
        putLength = keyWriteResult.getValue();
      } else {
        try (OzoneOutputStream output = getClientProtocol().createKey(
            volume.getName(), bucketName, keyPath, length, replicationConfig,
            customMetadata, tags)) {
          long metadataLatencyNs =
              getMetrics().updatePutKeyMetadataStats(startNanos);
          perf.appendMetaLatencyNanos(metadataLatencyNs);
          putLength = IOUtils.copyLarge(digestInputStream, output, 0, length,
              new byte[getIOBufferSize(length)]);
          eTag = DatatypeConverter.printHexBinary(
                  digestInputStream.getMessageDigest().digest())
              .toLowerCase();
          output.getMetadata().put(ETAG, eTag);
        }
      }
      getMetrics().incPutKeySuccessLength(putLength);
      perf.appendSizeBytes(putLength);
      return Response.ok()
          .header(ETAG, wrapInQuotes(eTag))
          .status(HttpStatus.SC_OK)
          .build();
    } catch (OMException ex) {
      auditSuccess = false;
      auditWriteFailure(s3GAction, ex);
      if (taggingMarker != null) {
        getMetrics().updatePutObjectTaggingFailureStats(startNanos);
      } else if (copyHeader != null) {
        getMetrics().updateCopyObjectFailureStats(startNanos);
      } else {
        getMetrics().updateCreateKeyFailureStats(startNanos);
      }
      if (ex.getResult() == ResultCodes.NOT_A_FILE) {
        OS3Exception os3Exception = newError(INVALID_REQUEST, keyPath, ex);
        os3Exception.setErrorMessage("An error occurred (InvalidRequest) " +
            "when calling the PutObject/MPU PartUpload operation: " +
            OmConfig.Keys.ENABLE_FILESYSTEM_PATHS + " is enabled Keys are" +
            " considered as Unix Paths. Path has Violated FS Semantics " +
            "which caused put operation to fail.");
        throw os3Exception;
      } else if (isAccessDenied(ex)) {
        throw newError(S3ErrorTable.ACCESS_DENIED, keyPath, ex);
      } else if (ex.getResult() == ResultCodes.QUOTA_EXCEEDED) {
        throw newError(S3ErrorTable.QUOTA_EXCEEDED, keyPath, ex);
      } else if (ex.getResult() == ResultCodes.BUCKET_NOT_FOUND) {
        throw newError(S3ErrorTable.NO_SUCH_BUCKET, bucketName, ex);
      } else if (ex.getResult() == ResultCodes.FILE_ALREADY_EXISTS) {
        throw newError(S3ErrorTable.NO_OVERWRITE, keyPath, ex);
      }
      throw ex;
    } catch (Exception ex) {
      auditSuccess = false;
      auditWriteFailure(s3GAction, ex);
      if (aclMarker != null) {
        getMetrics().updatePutObjectAclFailureStats(startNanos);
      } else if (taggingMarker != null) {
        getMetrics().updatePutObjectTaggingFailureStats(startNanos);
      } else if (copyHeader != null) {
        getMetrics().updateCopyObjectFailureStats(startNanos);
      } else {
        getMetrics().updateCreateKeyFailureStats(startNanos);
      }
      throw ex;
    } finally {
      // Reset the thread-local message digest instance in case of exception
      // and MessageDigest#digest is never called
      if (digestInputStream != null) {
        digestInputStream.getMessageDigest().reset();
      }
      if (auditSuccess) {
        long opLatencyNs = getMetrics().updateCreateKeySuccessStats(startNanos);
        perf.appendOpLatencyNanos(opLatencyNs);
        AUDIT.logWriteSuccess(buildAuditMessageForSuccess(s3GAction,
            getAuditParameters(), perf));
      }
    }
  }

  /**
   * Rest endpoint to download object from a bucket, if query param uploadId
   * is specified, request for list parts of a multipart upload key with
   * specific uploadId.
   * <p>
   * See: https://docs.aws.amazon.com/AmazonS3/latest/API/RESTObjectGET.html
   * https://docs.aws.amazon.com/AmazonS3/latest/API/mpUploadListParts.html
   * for more details.
   */
  @SuppressWarnings({"checkstyle:MethodLength", "checkstyle:ParameterNumber"})
  @GET
  public Response get(
      @PathParam("bucket") String bucketName,
      @PathParam("path") String keyPath,
      @QueryParam("partNumber") int partNumber,
      @QueryParam("uploadId") String uploadId,
      @QueryParam("max-parts") @DefaultValue("1000") int maxParts,
      @QueryParam("part-number-marker") String partNumberMarker,
      @QueryParam("tagging") String taggingMarker)
      throws IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();
    S3GAction s3GAction = S3GAction.GET_KEY;
    PerformanceStringBuilder perf = new PerformanceStringBuilder();
    try {
      if (taggingMarker != null) {
        s3GAction = S3GAction.GET_OBJECT_TAGGING;
        return getObjectTagging(bucketName, keyPath);
      }

      if (uploadId != null) {
        // When we have uploadId, this is the request for list Parts.
        s3GAction = S3GAction.LIST_PARTS;
        int partMarker = parsePartNumberMarker(partNumberMarker);
        Response response = listParts(bucketName, keyPath, uploadId,
            partMarker, maxParts, perf);
        AUDIT.logReadSuccess(buildAuditMessageForSuccess(s3GAction,
            getAuditParameters(), perf));
        return response;
      }

      OzoneKeyDetails keyDetails = (partNumber != 0) ?
          getClientProtocol().getS3KeyDetails(bucketName, keyPath, partNumber) :
          getClientProtocol().getS3KeyDetails(bucketName, keyPath);

      isFile(keyPath, keyDetails);

      long length = keyDetails.getDataSize();

      LOG.debug("Data length of the key {} is {}", keyPath, length);

      String rangeHeaderVal = headers.getHeaderString(RANGE_HEADER);
      RangeHeader rangeHeader = null;

      LOG.debug("range Header provided value: {}", rangeHeaderVal);

      if (rangeHeaderVal != null) {
        rangeHeader = RangeHeaderParserUtil.parseRangeHeader(rangeHeaderVal,
            length);
        LOG.debug("range Header provided: {}", rangeHeader);
        if (rangeHeader.isInValidRange()) {
          throw newError(S3ErrorTable.INVALID_RANGE, rangeHeaderVal);
        }
      }
      ResponseBuilder responseBuilder;

      if (rangeHeaderVal == null || rangeHeader.isReadFull()) {
        StreamingOutput output = dest -> {
          try (OzoneInputStream key = keyDetails.getContent()) {
            long readLength = IOUtils.copy(key, dest, getIOBufferSize(keyDetails.getDataSize()));
            getMetrics().incGetKeySuccessLength(readLength);
            perf.appendSizeBytes(readLength);
          }
          long opLatencyNs =  getMetrics().updateGetKeySuccessStats(startNanos);
          perf.appendOpLatencyNanos(opLatencyNs);
          AUDIT.logReadSuccess(buildAuditMessageForSuccess(S3GAction.GET_KEY,
              getAuditParameters(), perf));
        };
        responseBuilder = Response
            .ok(output)
            .header(CONTENT_LENGTH, keyDetails.getDataSize());

      } else {

        long startOffset = rangeHeader.getStartOffset();
        long endOffset = rangeHeader.getEndOffset();
        // eg. if range header is given as bytes=0-0, then we should return 1
        // byte from start offset
        long copyLength = endOffset - startOffset + 1;
        StreamingOutput output = dest -> {
          try (OzoneInputStream ozoneInputStream = keyDetails.getContent()) {
            ozoneInputStream.seek(startOffset);
            long readLength = IOUtils.copyLarge(ozoneInputStream, dest, 0,
                copyLength, new byte[getIOBufferSize(copyLength)]);
            getMetrics().incGetKeySuccessLength(readLength);
            perf.appendSizeBytes(readLength);
          }
          long opLatencyNs = getMetrics().updateGetKeySuccessStats(startNanos);
          perf.appendOpLatencyNanos(opLatencyNs);
          AUDIT.logReadSuccess(buildAuditMessageForSuccess(S3GAction.GET_KEY,
              getAuditParameters(), perf));
        };
        responseBuilder = Response
            .status(Status.PARTIAL_CONTENT)
            .entity(output)
            .header(CONTENT_LENGTH, copyLength);

        String contentRangeVal = RANGE_HEADER_SUPPORTED_UNIT + " " +
            rangeHeader.getStartOffset() + "-" + rangeHeader.getEndOffset() +
            "/" + length;

        responseBuilder.header(CONTENT_RANGE_HEADER, contentRangeVal);
      }
      responseBuilder
          .header(ACCEPT_RANGE_HEADER, RANGE_HEADER_SUPPORTED_UNIT);

      String eTag = keyDetails.getMetadata().get(ETAG);
      if (eTag != null) {
        responseBuilder.header(ETAG, wrapInQuotes(eTag));
        String partsCount = extractPartsCount(eTag);
        if (partsCount != null) {
          responseBuilder.header(MP_PARTS_COUNT, partsCount);
        }
      }

      // if multiple query parameters having same name,
      // Only the first parameters will be recognized
      // eg:
      // http://localhost:9878/bucket/key?response-expires=1&response-expires=2
      // only response-expires=1 is valid
      MultivaluedMap<String, String> queryParams = context
          .getUriInfo().getQueryParameters();

      for (Map.Entry<String, String> entry :
          overrideQueryParameter.entrySet()) {
        String headerValue = headers.getHeaderString(entry.getKey());

        /* "Overriding Response Header" by query parameter, See:
        https://docs.aws.amazon.com/de_de/AmazonS3/latest/API/API_GetObject.html
        */
        String queryValue = queryParams.getFirst(entry.getValue());
        if (queryValue != null) {
          headerValue = queryValue;
        }
        if (headerValue != null) {
          responseBuilder.header(entry.getKey(), headerValue);
        }
      }
      addLastModifiedDate(responseBuilder, keyDetails);
      addTagCountIfAny(responseBuilder, keyDetails);
      long metadataLatencyNs =
          getMetrics().updateGetKeyMetadataStats(startNanos);
      perf.appendMetaLatencyNanos(metadataLatencyNs);
      return responseBuilder.build();
    } catch (OMException ex) {
      AUDIT.logReadFailure(
          buildAuditMessageForFailure(s3GAction, getAuditParameters(), ex)
      );
      if (taggingMarker != null) {
        getMetrics().updateGetObjectTaggingFailureStats(startNanos);
      } else if (uploadId != null) {
        getMetrics().updateListPartsFailureStats(startNanos);
      } else {
        getMetrics().updateGetKeyFailureStats(startNanos);
      }
      if (ex.getResult() == ResultCodes.KEY_NOT_FOUND) {
        throw newError(S3ErrorTable.NO_SUCH_KEY, keyPath, ex);
      } else if (isAccessDenied(ex)) {
        throw newError(S3ErrorTable.ACCESS_DENIED, keyPath, ex);
      } else if (ex.getResult() == ResultCodes.BUCKET_NOT_FOUND) {
        throw newError(S3ErrorTable.NO_SUCH_BUCKET, bucketName, ex);
      } else {
        throw ex;
      }
    } catch (Exception ex) {
      AUDIT.logReadFailure(
          buildAuditMessageForFailure(s3GAction, getAuditParameters(), ex)
      );
      throw ex;
    }
  }

  static void addLastModifiedDate(
      ResponseBuilder responseBuilder, OzoneKey key) {

    ZonedDateTime lastModificationTime = key.getModificationTime()
        .atZone(ZoneId.of(OzoneConsts.OZONE_TIME_ZONE));

    responseBuilder
        .header(LAST_MODIFIED,
            RFC1123Util.FORMAT.format(lastModificationTime));
  }

  static void addTagCountIfAny(
      ResponseBuilder responseBuilder, OzoneKey key) {
    // See x-amz-tagging-count in https://docs.aws.amazon.com/AmazonS3/latest/API/API_GetObject.html
    // The number of tags, IF ANY, on the object, when you have the relevant
    // permission to read object tags
    if (!key.getTags().isEmpty()) {
      responseBuilder
          .header(TAG_COUNT_HEADER, key.getTags().size());
    }
  }

  /**
   * Rest endpoint to check existence of an object in a bucket.
   * <p>
   * See: https://docs.aws.amazon.com/AmazonS3/latest/API/RESTObjectHEAD.html
   * for more details.
   */
  @HEAD
  public Response head(
      @PathParam("bucket") String bucketName,
      @PathParam("path") String keyPath) throws IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();
    S3GAction s3GAction = S3GAction.HEAD_KEY;

    OzoneKey key;
    try {
      key = getClientProtocol().headS3Object(bucketName, keyPath);

      isFile(keyPath, key);
      // TODO: return the specified range bytes of this object.
    } catch (OMException ex) {
      AUDIT.logReadFailure(
          buildAuditMessageForFailure(s3GAction, getAuditParameters(), ex));
      getMetrics().updateHeadKeyFailureStats(startNanos);
      if (ex.getResult() == ResultCodes.KEY_NOT_FOUND) {
        // Just return 404 with no content
        return Response.status(Status.NOT_FOUND).build();
      } else if (isAccessDenied(ex)) {
        throw newError(S3ErrorTable.ACCESS_DENIED, keyPath, ex);
      } else if (ex.getResult() == ResultCodes.BUCKET_NOT_FOUND) {
        throw newError(S3ErrorTable.NO_SUCH_BUCKET, bucketName, ex);
      } else {
        throw ex;
      }
    } catch (Exception ex) {
      AUDIT.logReadFailure(
          buildAuditMessageForFailure(s3GAction, getAuditParameters(), ex));
      throw ex;
    }

    ResponseBuilder response = Response.ok().status(HttpStatus.SC_OK)
        .header("Content-Length", key.getDataSize())
        .header("Content-Type", "binary/octet-stream");

    String eTag = key.getMetadata().get(ETAG);
    if (eTag != null) {
      // Should not return ETag header if the ETag is not set
      // doing so will result in "null" string being returned instead
      // which breaks some AWS SDK implementation
      response.header(ETAG, wrapInQuotes(eTag));
      String partsCount = extractPartsCount(eTag);
      if (partsCount != null) {
        response.header(MP_PARTS_COUNT, partsCount);
      }
    }

    addLastModifiedDate(response, key);
    addCustomMetadataHeaders(response, key);
    getMetrics().updateHeadKeySuccessStats(startNanos);
    AUDIT.logReadSuccess(buildAuditMessageForSuccess(s3GAction,
        getAuditParameters()));
    return response.build();
  }

  private void isFile(String keyPath, OzoneKey key) throws OMException {
    /*
      Necessary for directories in buckets with FSO layout.
      Intended for apps which use Hadoop S3A.
      Example of such app is Trino (through Hive connector).
     */
    boolean isFsoDirCreationEnabled = ozoneConfiguration
        .getBoolean(OZONE_S3G_FSO_DIRECTORY_CREATION_ENABLED,
            OZONE_S3G_FSO_DIRECTORY_CREATION_ENABLED_DEFAULT);
    if (isFsoDirCreationEnabled &&
        !key.isFile() &&
        !keyPath.endsWith("/")) {
      throw new OMException(ResultCodes.KEY_NOT_FOUND);
    }
  }

  /**
   * Abort multipart upload request.
   * @param bucket
   * @param key
   * @param uploadId
   * @return Response
   * @throws IOException
   * @throws OS3Exception
   */
  private Response abortMultipartUpload(OzoneVolume volume, String bucket,
                                        String key, String uploadId)
      throws IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();
    try {
      getClientProtocol().abortMultipartUpload(volume.getName(), bucket,
          key, uploadId);
    } catch (OMException ex) {
      if (ex.getResult() == ResultCodes.NO_SUCH_MULTIPART_UPLOAD_ERROR) {
        throw newError(S3ErrorTable.NO_SUCH_UPLOAD, uploadId, ex);
      } else if (ex.getResult() == ResultCodes.BUCKET_NOT_FOUND) {
        throw newError(S3ErrorTable.NO_SUCH_BUCKET, bucket, ex);
      }
      throw ex;
    }
    getMetrics().updateAbortMultipartUploadSuccessStats(startNanos);
    return Response
        .status(Status.NO_CONTENT)
        .build();
  }


  /**
   * Delete a specific object from a bucket, if query param uploadId is
   * specified, this request is for abort multipart upload.
   * <p>
   * See: https://docs.aws.amazon.com/AmazonS3/latest/API/RESTObjectDELETE.html
   * https://docs.aws.amazon.com/AmazonS3/latest/API/mpUploadAbort.html
   * for more details.
   */
  @DELETE
  @SuppressWarnings("emptyblock")
  public Response delete(
      @PathParam("bucket") String bucketName,
      @PathParam("path") String keyPath,
      @QueryParam("uploadId") @DefaultValue("") String uploadId,
      @QueryParam("tagging") String taggingMarker) throws
      IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();
    S3GAction s3GAction = S3GAction.DELETE_KEY;

    try {
      OzoneVolume volume = getVolume();
      if (taggingMarker != null) {
        s3GAction = S3GAction.DELETE_OBJECT_TAGGING;
        return deleteObjectTagging(volume, bucketName, keyPath);
      }

      if (uploadId != null && !uploadId.equals("")) {
        s3GAction = S3GAction.ABORT_MULTIPART_UPLOAD;
        return abortMultipartUpload(volume, bucketName, keyPath, uploadId);
      }
      getClientProtocol().deleteKey(volume.getName(), bucketName,
          keyPath, false);
    } catch (OMException ex) {
      AUDIT.logWriteFailure(
          buildAuditMessageForFailure(s3GAction, getAuditParameters(), ex));
      if (uploadId != null && !uploadId.equals("")) {
        getMetrics().updateAbortMultipartUploadFailureStats(startNanos);
      } else {
        getMetrics().updateDeleteKeyFailureStats(startNanos);
      }
      if (ex.getResult() == ResultCodes.BUCKET_NOT_FOUND) {
        throw newError(S3ErrorTable.NO_SUCH_BUCKET, bucketName, ex);
      } else if (ex.getResult() == ResultCodes.KEY_NOT_FOUND) {
        //NOT_FOUND is not a problem, AWS doesn't throw exception for missing
        // keys. Just return 204
      } else if (ex.getResult() == ResultCodes.DIRECTORY_NOT_EMPTY) {
        // With PREFIX metadata layout, a dir deletion without recursive flag
        // to true will throw DIRECTORY_NOT_EMPTY error for a non-empty dir.
        // NOT_FOUND is not a problem, AWS doesn't throw exception for missing
        // keys. Just return 204
      } else if (isAccessDenied(ex)) {
        throw newError(S3ErrorTable.ACCESS_DENIED, keyPath, ex);
      } else if (ex.getResult() == ResultCodes.NOT_SUPPORTED_OPERATION) {
        // When deleteObjectTagging operation is applied on FSO directory
        throw S3ErrorTable.newError(S3ErrorTable.NOT_IMPLEMENTED, keyPath);
      } else {
        throw ex;
      }
    } catch (Exception ex) {
      AUDIT.logWriteFailure(
          buildAuditMessageForFailure(s3GAction, getAuditParameters(), ex));
      if (taggingMarker != null) {
        getMetrics().updateDeleteObjectTaggingFailureStats(startNanos);
      } else if (uploadId != null && !uploadId.equals("")) {
        getMetrics().updateAbortMultipartUploadFailureStats(startNanos);
      } else {
        getMetrics().updateDeleteKeyFailureStats(startNanos);
      }
      throw ex;
    }
    getMetrics().updateDeleteKeySuccessStats(startNanos);
    AUDIT.logWriteSuccess(buildAuditMessageForSuccess(s3GAction,
        getAuditParameters()));
    return Response
        .status(Status.NO_CONTENT)
        .build();
  }

  /**
   * Initialize MultiPartUpload request.
   * <p>
   * Note: the specific content type is set by the HeaderPreprocessor.
   */
  @POST
  @Produces(MediaType.APPLICATION_XML)
  @Consumes(HeaderPreprocessor.MULTIPART_UPLOAD_MARKER)
  public Response initializeMultipartUpload(
      @PathParam("bucket") String bucket,
      @PathParam("path") String key
  )
      throws IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();
    S3GAction s3GAction = S3GAction.INIT_MULTIPART_UPLOAD;

    try {
      OzoneBucket ozoneBucket = getBucket(bucket);
      String storageType = headers.getHeaderString(STORAGE_CLASS_HEADER);

      Map<String, String> customMetadata =
          getCustomMetadataFromHeaders(headers.getRequestHeaders());

      Map<String, String> tags = getTaggingFromHeaders(headers);

      ReplicationConfig replicationConfig =
          getReplicationConfig(ozoneBucket, storageType);

      OmMultipartInfo multipartInfo =
          ozoneBucket.initiateMultipartUpload(key, replicationConfig, customMetadata, tags);

      MultipartUploadInitiateResponse multipartUploadInitiateResponse = new
          MultipartUploadInitiateResponse();

      multipartUploadInitiateResponse.setBucket(bucket);
      multipartUploadInitiateResponse.setKey(key);
      multipartUploadInitiateResponse.setUploadID(multipartInfo.getUploadID());

      AUDIT.logWriteSuccess(
          buildAuditMessageForSuccess(s3GAction, getAuditParameters()));
      getMetrics().updateInitMultipartUploadSuccessStats(startNanos);
      return Response.status(Status.OK).entity(
          multipartUploadInitiateResponse).build();
    } catch (OMException ex) {
      auditWriteFailure(s3GAction, ex);
      getMetrics().updateInitMultipartUploadFailureStats(startNanos);
      if (isAccessDenied(ex)) {
        throw newError(S3ErrorTable.ACCESS_DENIED, key, ex);
      }
      throw ex;
    } catch (Exception ex) {
      AUDIT.logWriteFailure(
          buildAuditMessageForFailure(s3GAction, getAuditParameters(), ex));
      getMetrics().updateInitMultipartUploadFailureStats(startNanos);
      throw ex;
    }
  }

  private ReplicationConfig getReplicationConfig(OzoneBucket ozoneBucket,
      String storageType) throws OS3Exception {
    if (StringUtils.isEmpty(storageType)) {
      S3StorageType defaultStorageType = S3StorageType.getDefault(ozoneConfiguration);
      storageType = (defaultStorageType != null ? defaultStorageType.toString() : null);
    }

    ReplicationConfig clientConfiguredReplicationConfig = null;
    String replication = ozoneConfiguration.get(OZONE_REPLICATION);
    if (replication != null) {
      clientConfiguredReplicationConfig = ReplicationConfig.parse(
          ReplicationType.valueOf(ozoneConfiguration
              .get(OZONE_REPLICATION_TYPE, OZONE_REPLICATION_TYPE_DEFAULT)),
          replication, ozoneConfiguration);
    }
    return S3Utils.resolveS3ClientSideReplicationConfig(storageType,
        clientConfiguredReplicationConfig, ozoneBucket.getReplicationConfig());
  }

  /**
   * Complete a multipart upload.
   */
  @POST
  @Produces(MediaType.APPLICATION_XML)
  public Response completeMultipartUpload(@PathParam("bucket") String bucket,
      @PathParam("path") String key,
      @QueryParam("uploadId") @DefaultValue("") String uploadID,
      CompleteMultipartUploadRequest multipartUploadRequest)
      throws IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();
    S3GAction s3GAction = S3GAction.COMPLETE_MULTIPART_UPLOAD;
    OzoneVolume volume = getVolume();
    // Using LinkedHashMap to preserve ordering of parts list.
    Map<Integer, String> partsMap = new LinkedHashMap<>();
    List<CompleteMultipartUploadRequest.Part> partList =
        multipartUploadRequest.getPartList();

    OmMultipartUploadCompleteInfo omMultipartUploadCompleteInfo;
    try {
      for (CompleteMultipartUploadRequest.Part part : partList) {
        partsMap.put(part.getPartNumber(), part.getETag());
      }
      if (LOG.isDebugEnabled()) {
        LOG.debug("Parts map {}", partsMap);
      }

      omMultipartUploadCompleteInfo = getClientProtocol()
          .completeMultipartUpload(volume.getName(), bucket, key, uploadID,
              partsMap);
      CompleteMultipartUploadResponse completeMultipartUploadResponse =
          new CompleteMultipartUploadResponse();
      completeMultipartUploadResponse.setBucket(bucket);
      completeMultipartUploadResponse.setKey(key);
      completeMultipartUploadResponse.setETag(
          wrapInQuotes(omMultipartUploadCompleteInfo.getHash()));
      // Location also setting as bucket name.
      completeMultipartUploadResponse.setLocation(bucket);
      AUDIT.logWriteSuccess(
          buildAuditMessageForSuccess(s3GAction, getAuditParameters()));
      getMetrics().updateCompleteMultipartUploadSuccessStats(startNanos);
      return Response.status(Status.OK).entity(completeMultipartUploadResponse)
          .build();
    } catch (OMException ex) {
      auditWriteFailure(s3GAction, ex);
      getMetrics().updateCompleteMultipartUploadFailureStats(startNanos);
      if (ex.getResult() == ResultCodes.INVALID_PART) {
        throw newError(S3ErrorTable.INVALID_PART, key, ex);
      } else if (ex.getResult() == ResultCodes.INVALID_PART_ORDER) {
        throw newError(S3ErrorTable.INVALID_PART_ORDER, key, ex);
      } else if (ex.getResult() == ResultCodes.NO_SUCH_MULTIPART_UPLOAD_ERROR) {
        throw newError(NO_SUCH_UPLOAD, uploadID, ex);
      } else if (ex.getResult() == ResultCodes.ENTITY_TOO_SMALL) {
        throw newError(ENTITY_TOO_SMALL, key, ex);
      } else if (ex.getResult() == ResultCodes.INVALID_REQUEST) {
        OS3Exception os3Exception = newError(INVALID_REQUEST, key, ex);
        os3Exception.setErrorMessage("An error occurred (InvalidRequest) " +
            "when calling the CompleteMultipartUpload operation: You must " +
            "specify at least one part");
        throw os3Exception;
      } else if (ex.getResult() == ResultCodes.NOT_A_FILE) {
        OS3Exception os3Exception = newError(INVALID_REQUEST, key, ex);
        os3Exception.setErrorMessage("An error occurred (InvalidRequest) " +
            "when calling the CompleteMultipartUpload operation: " +
            OmConfig.Keys.ENABLE_FILESYSTEM_PATHS + " is enabled Keys are " +
            "considered as Unix Paths. A directory already exists with a " +
            "given KeyName caused failure for MPU");
        throw os3Exception;
      } else if (ex.getResult() == ResultCodes.BUCKET_NOT_FOUND) {
        throw newError(S3ErrorTable.NO_SUCH_BUCKET, bucket, ex);
      }
      throw ex;
    } catch (Exception ex) {
      auditWriteFailure(s3GAction, ex);
      throw ex;
    }
  }

  @SuppressWarnings({"checkstyle:MethodLength", "checkstyle:ParameterNumber"})
  private Response createMultipartKey(OzoneVolume volume, String bucket,
      String key, long length, int partNumber, String uploadID,
      final InputStream body, PerformanceStringBuilder perf)
      throws IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();
    String copyHeader = null;
    DigestInputStream digestInputStream = null;
    try {

      if (S3Utils.hasSignedPayloadHeader(headers)) {
        digestInputStream = new DigestInputStream(new SignedChunksInputStream(body),
            getMessageDigestInstance());
        length = Long.parseLong(
            headers.getHeaderString(DECODED_CONTENT_LENGTH_HEADER));
      } else {
        digestInputStream = new DigestInputStream(body, getMessageDigestInstance());
      }

      copyHeader = headers.getHeaderString(COPY_SOURCE_HEADER);
      String storageType = headers.getHeaderString(STORAGE_CLASS_HEADER);
      final OzoneBucket ozoneBucket = volume.getBucket(bucket);
      ReplicationConfig replicationConfig =
          getReplicationConfig(ozoneBucket, storageType);

      boolean enableEC = false;
      if ((replicationConfig != null &&
          replicationConfig.getReplicationType()  == EC) ||
          ozoneBucket.getReplicationConfig() instanceof ECReplicationConfig) {
        enableEC = true;
      }

      if (datastreamEnabled && !enableEC && copyHeader == null) {
        perf.appendStreamMode();
        return ObjectEndpointStreaming
            .createMultipartKey(ozoneBucket, key, length, partNumber,
                uploadID, chunkSize, digestInputStream, perf);
      }
      // OmMultipartCommitUploadPartInfo can only be gotten after the
      // OzoneOutputStream is closed, so we need to save the OzoneOutputStream
      final OzoneOutputStream outputStream;
      long metadataLatencyNs;
      if (copyHeader != null) {
        Pair<String, String> result = parseSourceHeader(copyHeader);
        String sourceBucket = result.getLeft();
        String sourceKey = result.getRight();

        OzoneKeyDetails sourceKeyDetails = getClientProtocol().getKeyDetails(
            volume.getName(), sourceBucket, sourceKey);
        String range =
            headers.getHeaderString(COPY_SOURCE_HEADER_RANGE);
        RangeHeader rangeHeader = null;
        if (range != null) {
          rangeHeader = RangeHeaderParserUtil.parseRangeHeader(range, 0);
          // When copy Range, the size of the target key is the
          // length specified by COPY_SOURCE_HEADER_RANGE.
          length = rangeHeader.getEndOffset() -
              rangeHeader.getStartOffset() + 1;
        } else {
          length = sourceKeyDetails.getDataSize();
        }
        Long sourceKeyModificationTime = sourceKeyDetails
            .getModificationTime().toEpochMilli();
        String copySourceIfModifiedSince =
            headers.getHeaderString(COPY_SOURCE_IF_MODIFIED_SINCE);
        String copySourceIfUnmodifiedSince =
            headers.getHeaderString(COPY_SOURCE_IF_UNMODIFIED_SINCE);
        if (!checkCopySourceModificationTime(sourceKeyModificationTime,
            copySourceIfModifiedSince, copySourceIfUnmodifiedSince)) {
          throw newError(PRECOND_FAILED, sourceBucket + "/" + sourceKey);
        }

        try (OzoneInputStream sourceObject = sourceKeyDetails.getContent()) {
          long copyLength;
          if (range != null) {
            final long skipped =
                sourceObject.skip(rangeHeader.getStartOffset());
            if (skipped != rangeHeader.getStartOffset()) {
              throw new EOFException(
                  "Bytes to skip: "
                      + rangeHeader.getStartOffset() + " actual: " + skipped);
            }
            try (OzoneOutputStream ozoneOutputStream = getClientProtocol()
                .createMultipartKey(volume.getName(), bucket, key, length,
                    partNumber, uploadID)) {
              metadataLatencyNs =
                  getMetrics().updateCopyKeyMetadataStats(startNanos);
              copyLength = IOUtils.copyLarge(
                  sourceObject, ozoneOutputStream, 0, length, new byte[getIOBufferSize(length)]);
              ozoneOutputStream.getMetadata()
                  .putAll(sourceKeyDetails.getMetadata());
              outputStream = ozoneOutputStream;
            }
          } else {
            try (OzoneOutputStream ozoneOutputStream = getClientProtocol()
                .createMultipartKey(volume.getName(), bucket, key, length,
                    partNumber, uploadID)) {
              metadataLatencyNs =
                  getMetrics().updateCopyKeyMetadataStats(startNanos);
              copyLength = IOUtils.copyLarge(sourceObject, ozoneOutputStream, 0, length,
                  new byte[getIOBufferSize(length)]);
              ozoneOutputStream.getMetadata()
                  .putAll(sourceKeyDetails.getMetadata());
              outputStream = ozoneOutputStream;
            }
          }
          getMetrics().incCopyObjectSuccessLength(copyLength);
          perf.appendSizeBytes(copyLength);
        }
      } else {
        long putLength;
        try (OzoneOutputStream ozoneOutputStream = getClientProtocol()
            .createMultipartKey(volume.getName(), bucket, key, length,
                partNumber, uploadID)) {
          metadataLatencyNs =
              getMetrics().updatePutKeyMetadataStats(startNanos);
          putLength = IOUtils.copyLarge(digestInputStream, ozoneOutputStream, 0, length,
              new byte[getIOBufferSize(length)]);
          byte[] digest = digestInputStream.getMessageDigest().digest();
          ozoneOutputStream.getMetadata()
              .put(ETAG, DatatypeConverter.printHexBinary(digest).toLowerCase());
          outputStream = ozoneOutputStream;
        }
        getMetrics().incPutKeySuccessLength(putLength);
        perf.appendSizeBytes(putLength);
      }
      perf.appendMetaLatencyNanos(metadataLatencyNs);

      OmMultipartCommitUploadPartInfo omMultipartCommitUploadPartInfo =
          outputStream.getCommitUploadPartInfo();
      String eTag = omMultipartCommitUploadPartInfo.getETag();
      // If the OmMultipartCommitUploadPartInfo does not contain eTag,
      // fall back to MPU part name for compatibility in case the (old) OM
      // does not return the eTag field
      if (StringUtils.isEmpty(eTag)) {
        eTag = omMultipartCommitUploadPartInfo.getPartName();
      }

      if (copyHeader != null) {
        getMetrics().updateCopyObjectSuccessStats(startNanos);
        return Response.ok(new CopyPartResult(eTag)).build();
      } else {
        getMetrics().updateCreateMultipartKeySuccessStats(startNanos);
        return Response.ok().header(ETAG, eTag).build();
      }

    } catch (OMException ex) {
      if (copyHeader != null) {
        getMetrics().updateCopyObjectFailureStats(startNanos);
      } else {
        getMetrics().updateCreateMultipartKeyFailureStats(startNanos);
      }
      if (ex.getResult() == ResultCodes.NO_SUCH_MULTIPART_UPLOAD_ERROR) {
        throw newError(NO_SUCH_UPLOAD, uploadID, ex);
      } else if (isAccessDenied(ex)) {
        throw newError(S3ErrorTable.ACCESS_DENIED, bucket + "/" + key, ex);
      } else if (ex.getResult() == ResultCodes.INVALID_PART) {
        OS3Exception os3Exception = newError(
            S3ErrorTable.INVALID_ARGUMENT, String.valueOf(partNumber), ex);
        os3Exception.setErrorMessage(ex.getMessage());
        throw os3Exception;
      }
      throw ex;
    } finally {
      // Reset the thread-local message digest instance in case of exception
      // and MessageDigest#digest is never called
      if (digestInputStream != null) {
        digestInputStream.getMessageDigest().reset();
      }
    }
  }

  /**
   * Returns response for the listParts request.
   * See: https://docs.aws.amazon.com/AmazonS3/latest/API/mpUploadListParts.html
   * @param bucket
   * @param key
   * @param uploadID
   * @param partNumberMarker
   * @param maxParts
   * @return
   * @throws IOException
   * @throws OS3Exception
   */
  private Response listParts(String bucket, String key, String uploadID,
      int partNumberMarker, int maxParts, PerformanceStringBuilder perf)
      throws IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();
    ListPartsResponse listPartsResponse = new ListPartsResponse();
    try {
      OzoneBucket ozoneBucket = getBucket(bucket);
      OzoneMultipartUploadPartListParts ozoneMultipartUploadPartListParts =
          ozoneBucket.listParts(key, uploadID, partNumberMarker, maxParts);
      listPartsResponse.setBucket(bucket);
      listPartsResponse.setKey(key);
      listPartsResponse.setUploadID(uploadID);
      listPartsResponse.setMaxParts(maxParts);
      listPartsResponse.setPartNumberMarker(partNumberMarker);
      listPartsResponse.setTruncated(false);

      listPartsResponse.setStorageClass(S3StorageType.fromReplicationConfig(
          ozoneMultipartUploadPartListParts.getReplicationConfig()).toString());

      if (ozoneMultipartUploadPartListParts.isTruncated()) {
        listPartsResponse.setTruncated(
            ozoneMultipartUploadPartListParts.isTruncated());
        listPartsResponse.setNextPartNumberMarker(
            ozoneMultipartUploadPartListParts.getNextPartNumberMarker());
      }

      ozoneMultipartUploadPartListParts.getPartInfoList().forEach(partInfo -> {
        ListPartsResponse.Part part = new ListPartsResponse.Part();
        part.setPartNumber(partInfo.getPartNumber());
        // If the ETag field does not exist, use MPU part name for backward
        // compatibility
        part.setETag(StringUtils.isNotEmpty(partInfo.getETag()) ?
            partInfo.getETag() : partInfo.getPartName());
        part.setSize(partInfo.getSize());
        part.setLastModified(Instant.ofEpochMilli(
            partInfo.getModificationTime()));
        listPartsResponse.addPart(part);
      });
    } catch (OMException ex) {
      getMetrics().updateListPartsFailureStats(startNanos);
      if (ex.getResult() == ResultCodes.NO_SUCH_MULTIPART_UPLOAD_ERROR) {
        throw newError(NO_SUCH_UPLOAD, uploadID, ex);
      } else if (isAccessDenied(ex)) {
        throw newError(S3ErrorTable.ACCESS_DENIED,
            bucket + "/" + key + "/" + uploadID, ex);
      }
      throw ex;
    }
    long opLatencyNs = getMetrics().updateListPartsSuccessStats(startNanos);
    perf.appendCount(listPartsResponse.getPartList().size());
    perf.appendOpLatencyNanos(opLatencyNs);
    return Response.status(Status.OK).entity(listPartsResponse).build();
  }

  @VisibleForTesting
  public void setHeaders(HttpHeaders headers) {
    this.headers = headers;
  }

  @VisibleForTesting
  public void setContext(ContainerRequestContext context) {
    this.context = context;
  }

  @SuppressWarnings("checkstyle:ParameterNumber")
  void copy(OzoneVolume volume, DigestInputStream src, long srcKeyLen,
      String destKey, String destBucket,
      ReplicationConfig replication,
      Map<String, String> metadata,
      PerformanceStringBuilder perf, long startNanos,
      Map<String, String> tags)
      throws IOException {
    long copyLength;
    if (datastreamEnabled && !(replication != null &&
        replication.getReplicationType() == EC) &&
        srcKeyLen > datastreamMinLength) {
      perf.appendStreamMode();
      copyLength = ObjectEndpointStreaming
          .copyKeyWithStream(volume.getBucket(destBucket), destKey, srcKeyLen,
              chunkSize, replication, metadata, src, perf, startNanos, tags);
    } else {
      try (OzoneOutputStream dest = getClientProtocol()
          .createKey(volume.getName(), destBucket, destKey, srcKeyLen,
              replication, metadata, tags)) {
        long metadataLatencyNs =
            getMetrics().updateCopyKeyMetadataStats(startNanos);
        perf.appendMetaLatencyNanos(metadataLatencyNs);
        copyLength = IOUtils.copyLarge(src, dest, 0, srcKeyLen, new byte[getIOBufferSize(srcKeyLen)]);
        String eTag = DatatypeConverter.printHexBinary(src.getMessageDigest().digest()).toLowerCase();
        dest.getMetadata().put(ETAG, eTag);
      }
    }
    getMetrics().incCopyObjectSuccessLength(copyLength);
    perf.appendSizeBytes(copyLength);
  }

  @SuppressWarnings("checkstyle:ParameterNumber")
  private CopyObjectResponse copyObject(OzoneVolume volume,
      String copyHeader, String destBucket, String destkey,
      ReplicationConfig replicationConfig, boolean storageTypeDefault,
      PerformanceStringBuilder perf)
      throws OS3Exception, IOException {
    long startNanos = Time.monotonicNowNanos();
    Pair<String, String> result = parseSourceHeader(copyHeader);

    String sourceBucket = result.getLeft();
    String sourceKey = result.getRight();
    DigestInputStream sourceDigestInputStream = null;
    try {
      OzoneKeyDetails sourceKeyDetails = getClientProtocol().getKeyDetails(
          volume.getName(), sourceBucket, sourceKey);
      // Checking whether we trying to copying to it self.
      if (sourceBucket.equals(destBucket) && sourceKey
          .equals(destkey)) {
        // When copying to same storage type when storage type is provided,
        // we should not throw exception, as aws cli checks if any of the
        // options like storage type are provided or not when source and
        // dest are given same
        if (storageTypeDefault) {
          OS3Exception ex = newError(S3ErrorTable.INVALID_REQUEST, copyHeader);
          ex.setErrorMessage("This copy request is illegal because it is " +
              "trying to copy an object to it self itself without changing " +
              "the object's metadata, storage class, website redirect " +
              "location or encryption attributes.");
          throw ex;
        } else {
          // TODO: Actually here we should change storage type, as ozone
          // still does not support this just returning dummy response
          // for now
          CopyObjectResponse copyObjectResponse = new CopyObjectResponse();
          copyObjectResponse.setETag(wrapInQuotes(sourceKeyDetails.getMetadata().get(ETAG)));
          copyObjectResponse.setLastModified(Instant.ofEpochMilli(
              Time.now()));
          return copyObjectResponse;
        }
      }
      long sourceKeyLen = sourceKeyDetails.getDataSize();

      // Object tagging in copyObject with tagging directive
      Map<String, String> tags;
      String tagCopyDirective = headers.getHeaderString(TAG_DIRECTIVE_HEADER);
      if (StringUtils.isEmpty(tagCopyDirective) || tagCopyDirective.equals(CopyDirective.COPY.name())) {
        // Tag-set will be copied from the source directly
        tags = sourceKeyDetails.getTags();
      } else if (tagCopyDirective.equals(CopyDirective.REPLACE.name())) {
        // Replace the tags with the tags from the request headers
        tags = getTaggingFromHeaders(headers);
      } else {
        OS3Exception ex = newError(INVALID_ARGUMENT, tagCopyDirective);
        ex.setErrorMessage("An error occurred (InvalidArgument) " +
            "when calling the CopyObject operation: " +
            "The tagging copy directive specified is invalid. Valid values are COPY or REPLACE.");
        throw ex;
      }

      // Custom metadata in copyObject with metadata directive
      Map<String, String> customMetadata;
      String metadataCopyDirective = headers.getHeaderString(CUSTOM_METADATA_COPY_DIRECTIVE_HEADER);
      if (StringUtils.isEmpty(metadataCopyDirective) || metadataCopyDirective.equals(CopyDirective.COPY.name())) {
        // The custom metadata will be copied from the source key
        customMetadata = sourceKeyDetails.getMetadata();
      } else if (metadataCopyDirective.equals(CopyDirective.REPLACE.name())) {
        // Replace the metadata with the metadata form the request headers
        customMetadata = getCustomMetadataFromHeaders(headers.getRequestHeaders());
      } else {
        OS3Exception ex = newError(INVALID_ARGUMENT, metadataCopyDirective);
        ex.setErrorMessage("An error occurred (InvalidArgument) " +
            "when calling the CopyObject operation: " +
            "The metadata copy directive specified is invalid. Valid values are COPY or REPLACE.");
        throw ex;
      }

      try (OzoneInputStream src = getClientProtocol().getKey(volume.getName(),
          sourceBucket, sourceKey)) {
        getMetrics().updateCopyKeyMetadataStats(startNanos);
        sourceDigestInputStream = new DigestInputStream(src, getMessageDigestInstance());
        copy(volume, sourceDigestInputStream, sourceKeyLen, destkey, destBucket, replicationConfig,
                customMetadata, perf, startNanos, tags);
      }

      final OzoneKeyDetails destKeyDetails = getClientProtocol().getKeyDetails(
          volume.getName(), destBucket, destkey);

      getMetrics().updateCopyObjectSuccessStats(startNanos);
      CopyObjectResponse copyObjectResponse = new CopyObjectResponse();
      copyObjectResponse.setETag(wrapInQuotes(destKeyDetails.getMetadata().get(ETAG)));
      copyObjectResponse.setLastModified(destKeyDetails.getModificationTime());
      return copyObjectResponse;
    } catch (OMException ex) {
      if (ex.getResult() == ResultCodes.KEY_NOT_FOUND) {
        throw newError(S3ErrorTable.NO_SUCH_KEY, sourceKey, ex);
      } else if (ex.getResult() == ResultCodes.BUCKET_NOT_FOUND) {
        throw newError(S3ErrorTable.NO_SUCH_BUCKET, sourceBucket, ex);
      } else if (isAccessDenied(ex)) {
        throw newError(S3ErrorTable.ACCESS_DENIED,
            destBucket + "/" + destkey, ex);
      }
      throw ex;
    } finally {
      // Reset the thread-local message digest instance in case of exception
      // and MessageDigest#digest is never called
      if (sourceDigestInputStream != null) {
        sourceDigestInputStream.getMessageDigest().reset();
      }
    }
  }

  /**
   * Parse the key and bucket name from copy header.
   */
  @VisibleForTesting
  public static Pair<String, String> parseSourceHeader(String copyHeader)
      throws OS3Exception {
    String header = copyHeader;
    if (header.startsWith("/")) {
      header = copyHeader.substring(1);
    }
    int pos = header.indexOf('/');
    if (pos == -1) {
      OS3Exception ex = newError(INVALID_ARGUMENT, header);
      ex.setErrorMessage("Copy Source must mention the source bucket and " +
          "key: sourcebucket/sourcekey");
      throw ex;
    }

    try {
      String bucket = header.substring(0, pos);
      String key = urlDecode(header.substring(pos + 1));
      return Pair.of(bucket, key);
    } catch (UnsupportedEncodingException e) {
      OS3Exception ex = newError(INVALID_ARGUMENT, header, e);
      ex.setErrorMessage("Copy Source header could not be url-decoded");
      throw ex;
    }
  }

  private static int parsePartNumberMarker(String partNumberMarker) {
    int partMarker = 0;
    if (partNumberMarker != null) {
      partMarker = Integer.parseInt(partNumberMarker);
    }
    return partMarker;
  }

  // Parses date string and return long representation. Returns an
  // empty if DateStr is null or invalid. Dates in the future are
  // considered invalid.
  private static OptionalLong parseAndValidateDate(String ozoneDateStr) {
    long ozoneDateInMs;
    if (ozoneDateStr == null) {
      return OptionalLong.empty();
    }
    try {
      ozoneDateInMs = OzoneUtils.formatDate(ozoneDateStr);
    } catch (ParseException e) {
      // if time not parseable, then return empty()
      return OptionalLong.empty();
    }

    long currentDate = System.currentTimeMillis();
    if (ozoneDateInMs <= currentDate) {
      return OptionalLong.of(ozoneDateInMs);
    } else {
      // dates in the future are invalid, so return empty()
      return OptionalLong.empty();
    }
  }

  public static boolean checkCopySourceModificationTime(
      Long lastModificationTime,
      String copySourceIfModifiedSinceStr,
      String copySourceIfUnmodifiedSinceStr) {
    long copySourceIfModifiedSince = Long.MIN_VALUE;
    long copySourceIfUnmodifiedSince = Long.MAX_VALUE;

    OptionalLong modifiedDate =
        parseAndValidateDate(copySourceIfModifiedSinceStr);
    if (modifiedDate.isPresent()) {
      copySourceIfModifiedSince = modifiedDate.getAsLong();
    }

    OptionalLong unmodifiedDate =
        parseAndValidateDate(copySourceIfUnmodifiedSinceStr);
    if (unmodifiedDate.isPresent()) {
      copySourceIfUnmodifiedSince = unmodifiedDate.getAsLong();
    }
    return (copySourceIfModifiedSince <= lastModificationTime) &&
        (lastModificationTime <= copySourceIfUnmodifiedSince);
  }

  private Response putObjectTagging(OzoneVolume volume, String bucketName, String keyName, InputStream body)
      throws IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();
    S3Tagging tagging = null;
    try {
      tagging = new PutTaggingUnmarshaller().readFrom(body);
      tagging.validate();
    } catch (Exception ex) {
      OS3Exception exception = S3ErrorTable.newError(S3ErrorTable.MALFORMED_XML, keyName);
      exception.setErrorMessage(exception.getErrorMessage() + ". " + ex.getMessage());
      throw exception;
    }

    Map<String, String> tags = validateAndGetTagging(
        tagging.getTagSet().getTags(), // Nullity check was done in previous parsing step
        Tag::getKey,
        Tag::getValue
    );

    try {
      volume.getBucket(bucketName).putObjectTagging(keyName, tags);
    } catch (OMException ex) {
      if (ex.getResult() == ResultCodes.INVALID_REQUEST) {
        throw S3ErrorTable.newError(S3ErrorTable.INVALID_REQUEST, keyName);
      } else if (ex.getResult() == ResultCodes.PERMISSION_DENIED) {
        throw S3ErrorTable.newError(S3ErrorTable.ACCESS_DENIED, keyName);
      } else if (ex.getResult() == ResultCodes.KEY_NOT_FOUND) {
        throw S3ErrorTable.newError(S3ErrorTable.NO_SUCH_KEY, keyName);
      } else if (ex.getResult() == ResultCodes.NOT_SUPPORTED_OPERATION) {
        // When putObjectTagging operation is applied on FSO directory
        throw S3ErrorTable.newError(S3ErrorTable.NOT_IMPLEMENTED, keyName);
      }
      throw ex;
    }
    getMetrics().updatePutObjectTaggingSuccessStats(startNanos);
    return Response.ok().build();
  }

  private Response getObjectTagging(String bucketName, String keyName) throws IOException {
    long startNanos = Time.monotonicNowNanos();

    OzoneVolume volume = getVolume();

    Map<String, String> tagMap = volume.getBucket(bucketName).getObjectTagging(keyName);

    getMetrics().updateGetObjectTaggingSuccessStats(startNanos);
    return Response.ok(S3Tagging.fromMap(tagMap), MediaType.APPLICATION_XML_TYPE).build();
  }

  private Response deleteObjectTagging(OzoneVolume volume, String bucketName, String keyName)
      throws IOException, OS3Exception {
    long startNanos = Time.monotonicNowNanos();

    try {
      volume.getBucket(bucketName).deleteObjectTagging(keyName);
    } catch (OMException ex) {
      // Unlike normal key deletion that ignores the key not found exception
      // DeleteObjectTagging should throw the exception if the key does not exist
      if (ex.getResult() == ResultCodes.KEY_NOT_FOUND) {
        throw S3ErrorTable.newError(S3ErrorTable.NO_SUCH_KEY, keyName);
      }
      throw ex;
    }

    getMetrics().updateDeleteObjectTaggingSuccessStats(startNanos);
    return Response.noContent().build();
  }


  @VisibleForTesting
  public void setOzoneConfiguration(OzoneConfiguration config) {
    this.ozoneConfiguration = config;
  }

  @VisibleForTesting
  public boolean isDatastreamEnabled() {
    return datastreamEnabled;
  }

  static String wrapInQuotes(String value) {
    return "\"" + value + "\"";
  }

  @VisibleForTesting
  public MessageDigest getMessageDigestInstance() {
    return E_TAG_PROVIDER.get();
  }

  private String extractPartsCount(String eTag) {
    if (eTag.contains("-")) {
      String[] parts = eTag.replace("\"", "").split("-");
      String lastPart = parts[parts.length - 1];
      return lastPart.matches("\\d+") ? lastPart : null;
    }
    return null;
  }

  private int getIOBufferSize(long fileLength) {
    if (bufferSize == 0) {
      // this is mainly for unit tests as init() will not be called in the unit tests
      LOG.warn("buffer size is set to {}", IOUtils.DEFAULT_BUFFER_SIZE);
      bufferSize = IOUtils.DEFAULT_BUFFER_SIZE;
    }
    if (fileLength == 0) {
      // for empty file
      return bufferSize;
    } else {
      return fileLength < bufferSize ? (int) fileLength : bufferSize;
    }
  }
}
